/*******************************************************************************
 *******************************************************************************
  @file
    embdz19121-spi-m.ino
 
    EMBDZ19121
 

  @date
    05.04.2020

  @company:
    Made Systems, LLC

  @summary:
    This applicaiton code is designed to illustrate how to use the SPI bus
    to communicate with the "M" series a la mods smart modules.

    This example used the GPIO expander to set the module address bits of the
    GPIO header and then transfer data via the SPI bus
    

    (c) 2020 Made Systems, LLC. All rights reserved
********************************************************************************
*******************************************************************************/
#include "Wire.h"
#include "SPI.h"



#define HDWR_SPI_CE0_PIN  15
#define HDWR_SPI_CE1_PIN  25



typedef union {
    struct {
        uint8_t           b[4];
    };
    uint32_t              val;
} SPI_MSG_t;



static  uint8_t           gpio0;          // port 0 mirror variable of the GPIO expander
static  uint8_t           gpio1;          // port 1 mirror variable of the GPIO expander

        SPIClass         *spi             = NULL;
                                          // set the SPI clock to 1 (Mbps)
static  const  int        spiClk          = 1000000;
static  SPI_MSG_t         spiRxBfr;       // SPI bus receive data buffer
static  SPI_MSG_t         spiTxBfr;       // SPI bus transmit data buffer


    // initialize the hardware GPIO Expander and Serial UART
void setup() {
    Serial.begin(115200);
                    // setup port mirror variables
    gpio0                 = 0b00000000;
    gpio1                 = 0b00000000;

                    // setup the GPIO expander ports on EMBDZ19121
                    // port 0
                    //  - bit 0   = input   GPIO23  (header pin 16)
                    //  - bit 1   = input   GPIO24  (header pin 18)
                    //  - bit 2   = input   GPIO05  (header pin 29)
                    //  - bit 3   = input   GPIO06  (header pin 31)
                    //  - bit 4   = input   GPIO22  (header pin 15)
                    //  - bit 5   = input   GPIO27  (header pin 13)
                    //  - bit 6   = input   GPIO17  (header pin 11)
                    //  - bit 7   = input   GPIO04  (header pin 7)
                    // port 1
                    //  - bit 0   = output  (unused set low)
                    //  - bit 1   = output  (unused set low)
                    //  - bit 2   = output  (unused set low)
                    //  - bit 3   = output  (unused set low)
                    //  - bit 4   = input   GPIO26  (header pin 37)
                    //  - bit 5   = output  GPIO25  (header pin 22) Status LED
                    //  - bit 6   = input   GPIO16  (header pin 36)
                    //  - bit 7   = input   GPIO21  (header pin 38)

                    // setup the I2C bus
    Wire.begin(32, 33, 100000);
                    // initialize GPIO expander port 0 direction register
    Wire.beginTransmission(0x20);
    Wire.write(0x06);     // internal GPIO expander port 0 direction register address
                          // set the M series address bits (0-3) to outputs
    Wire.write(0xf0);
    Wire.endTransmission();
                    // initialize GPIO expander port 1 direction register
    Wire.beginTransmission(0x20);
    Wire.write(0x07);     // internal GPIO expander port 1 direction register address
    Wire.write(0xd0);
    Wire.endTransmission();
                    // initialize GPIO expander port 0 output values
                    // set the M series module address bits value to all 1's
                    // the address logic is inverted so 0b1111 = address 0
    gpio0                                 = gpio0 | 0b00001111;
    Wire.beginTransmission(0x20);
    Wire.write(0x02);     // internal GPIO expander port 0 output register address
    Wire.write(gpio0);
    Wire.endTransmission();
                    // initialize GPIO expander port 1 output values
    Wire.beginTransmission(0x20);
    Wire.write(0x03);     // internal GPIO expander port 1 output register address
    Wire.write(gpio1);
    Wire.endTransmission();

    Serial.println("GPIO Expander setup complete...");

                    // setup the system SPI bus
                    //
                    // setup pin of the ESP32 as the CE0 select line - initialize
                    // the pin high to put CE0 into an inactive state
    digitalWrite(HDWR_SPI_CE0_PIN, HIGH);
    pinMode(HDWR_SPI_CE0_PIN, OUTPUT);
                    // setup pin of the ESP32 as the CE1 select line - initialize
                    // the pin high to put CE1 into an inactive state
    digitalWrite(HDWR_SPI_CE1_PIN, HIGH);
    pinMode(HDWR_SPI_CE1_PIN, OUTPUT);
                    // instantiate the spi object and set the appropriate pins
                    // on the ESP32 for the EMBDZ19121
                    // ESP32 GPIO14 - CLK
                    // ESP32 GPIO12 - MISO
                    // ESP32 GPIO13 - MOSI
                    // ESP32 GPIO15 - CE0
    spi                                         = new SPIClass(HSPI);
    spi->begin(14,12,13,15);

    Serial.println("SPI bus setup complete...");
}


    // loop forever toggling the Status LED on/off and setting the LED on 
    // the a la mods "M" series smart module
void loop() {
                    // turn the EMBDZ19121 Status LED on
                    // setup the gpio1 mirror register to set bit 5 high without
                    // changing the other bits in the gpio1 mirror register
    gpio1                 = gpio1 | 0b00100000;
                    // write the port value to the GPIO Expander
    Wire.beginTransmission(0x20);
    Wire.write(0x03);     // internal GPIO expander port 1 output register address
    Wire.write(gpio1);
    Wire.endTransmission();

    Serial.println("Set Status LED on...");

                    // turn the "M" series module LED on
                    //
                    // The GPIO header module address lines were setup in the
                    // initialization of the GPIO Expander port 0 (see setup above)
                    //
                    // set up the SPI command to send to the "M" series module
    spiTxBfr.b[3]         = 0x02;     // register write command
    spiTxBfr.b[2]         = 0x90;     // internal regsiter to write to
    spiTxBfr.b[1]         = 0x80;     // set the LED control bit of register 0x90
    spiTxBfr.b[0]         = 0x05;     // set the RGB led bits to turn on color
                                      // bit 0  = RED
                                      // bit 1  = GRN
                                      // bit 2  = BLU
                    // set the SPI bus mode and clock
    spi->beginTransaction(SPISettings(spiClk, MSBFIRST, SPI_MODE0));
                    // set the CE0 select line active
    digitalWrite(HDWR_SPI_CE0_PIN, LOW);
                    // transfer the SPI a la mods command (4 bytes)
    spiRxBfr.b[3]         = spi->transfer(spiTxBfr.b[3]);
    spiRxBfr.b[2]         = spi->transfer(spiTxBfr.b[2]);
    spiRxBfr.b[1]         = spi->transfer(spiTxBfr.b[1]);
    spiRxBfr.b[0]         = spi->transfer(spiTxBfr.b[0]);
                    // set the CE0 select line inactive
    digitalWrite(HDWR_SPI_CE0_PIN, HIGH);
    spi->endTransaction();
    
    Serial.println("Set Module LED on...");

    delay(500);           // wait for 500 mSec

                    // turn the EMBDZ19121 Status LED off
                    // setup the gpio1 mirror register to set bit 5 low without
                    // changing the other bits in the gpio1 mirror register
    gpio1                 = gpio1 & 0b11011111;
                    // write the port value to the GPIO Expander
    Wire.beginTransmission(0x20);
    Wire.write(0x03);     // internal GPIO expander port 1 output register address
    Wire.write(gpio1);
    Wire.endTransmission();

    Serial.println("Set Status LED off...");

                    // turn the "M" series module LED off
                    //
                    // The GPIO header module address lines were setup in the
                    // initialization of the GPIO Expander port 0 (see setup above)
                    //
                    // set up the SPI command to send to the "M" series module
    spiTxBfr.b[3]         = 0x02;     // register write command
    spiTxBfr.b[2]         = 0x90;     // internal regsiter to write to
    spiTxBfr.b[1]         = 0x80;     // set the LED control bit of register 0x90
    spiTxBfr.b[0]         = 0x00;     // set the RGB led bits to turn off
                                      // bit 0  = RED
                                      // bit 1  = GRN
                                      // bit 2  = BLU
                    // set the SPI bus mode and clock
    spi->beginTransaction(SPISettings(spiClk, MSBFIRST, SPI_MODE0));
                    // set the CE0 select line active
    digitalWrite(HDWR_SPI_CE0_PIN, LOW);
                    // transfer the SPI a la mods command (4 bytes)
    spiRxBfr.b[3]         = spi->transfer(spiTxBfr.b[3]);
    spiRxBfr.b[2]         = spi->transfer(spiTxBfr.b[2]);
    spiRxBfr.b[1]         = spi->transfer(spiTxBfr.b[1]);
    spiRxBfr.b[0]         = spi->transfer(spiTxBfr.b[0]);
                    // set the CE0 select line inactive
    digitalWrite(HDWR_SPI_CE0_PIN, HIGH);
    spi->endTransaction();
    
    Serial.println("Set Module LED off...");

    delay(500);           // wait for 500 mSec
}
